--Autoren:
--Tobias Möllmann, inf9407
--Felix-Alexander Döppers inf7340

import List
    ( (\\)
    )

import System
    ( getArgs
    )

-- data types for the sudoku solver

-- matrix
type Matrix a = [Row a]
type Row a    = [a]

-- 25 x 25 grid with Digits

type Grid    = Matrix Digit
type Digit   = Char

type Choices = [Digit]

-- digits: ['A'-'Y'] and '-' as blank
digits :: [Digit]
digits = ['A'..'Y']

blank :: Digit -> Bool
blank  = (== '-')


prune :: Matrix Choices -> Matrix Choices
prune = pruneBy boxs . pruneBy cols . pruneBy rows
        where pruneBy f = f . map pruneRow . f

pruneRow :: Row Choices -> Row Choices
pruneRow row = map (remove fixed) row
               where fixed = [d | [d] <- row]

isSingleton :: [a] -> Bool
isSingleton [_] = True
isSingleton _ = False

remove xs ds = if isSingleton ds then ds else ds \\ xs

choices :: Grid -> Matrix Choices
choices = map (map choice)
  where choice d = if blank d then digits else [d]

expand :: Matrix Choices -> [Grid]
expand = cp . map cp

--cartesian product
cp            :: [[a]] -> [[a]]
cp []         = [[]]
cp (xs : xss) = [x : ys | x <- xs, ys <- cp xss]

valid   :: Grid -> Bool
valid g = all nodups (rows g) &&
          all nodups (cols g) &&
          all nodups (boxs g)

nodups          :: Eq a => [a] -> Bool
nodups []       = True
nodups (x : xs) = all (/= x) xs && nodups xs

rows :: Matrix a -> Matrix a
rows = id

cols            :: Matrix a -> Matrix a
cols [xs]       = [[x]| x <- xs]
cols (xs : xss) = zipWith (:) xs (cols xss)

boxs :: Matrix a -> Matrix a
boxs = map ungroup . ungroup . map cols . group . map group

group    :: [a] -> [[a]]
group [] = []
group xs = take 5 xs : group (drop 5 xs)

ungroup :: [[a]] -> [a]
ungroup = concat

expand1       :: Matrix Choices -> [Matrix Choices]
expand1 rows  = [rows1 ++ [row1 ++ [c] : row2] ++ rows2 | c <- cs]
              where 
              (rows1, row : rows2)  = break (any smallest) rows
              (row1, cs : row2)     = break smallest row
              smallest cs           = length cs == n
              n                     = minimum (counts rows)
              counts                = filter (/= 1) . (map length) . concat

complete              :: Matrix Choices -> Bool
complete              =  all (all isSingleton)

safe                  :: Matrix Choices -> Bool
safe m                =  all consistent (rows m) &&
                         all consistent (cols m) &&
                         all consistent (boxs m)

consistent            :: Row Choices -> Bool
consistent            =  nodups . concat . filter isSingleton

search           :: Matrix Choices -> [Grid]
search m
  | not (safe m) =  []
  | complete m'  =  [map (map head) m']
  | otherwise    =  concat (map search (expand1 m'))
  where m'       = prune m

-- solver algorithm
solve :: Grid -> [Grid]
solve = search . choices


readSudoku :: String -> Grid
readSudoku xs = toCols (suRows xs)
              where
              toCols (xs : xss) = (suCols xs) : (toCols xss)
              toCols [] = []
              suRows xs = snd $ unzip $ zip [1..25] (lines xs)
              suCols ys = map head (snd $ unzip $ zip [1..25] (words ys))

showSudoku     :: Grid -> IO()
showSudoku xss = putStr (unlines (insSpaces xss))

insSpaces :: [String] -> [String]
insSpaces (xs : xss) = (unwords (map (\x -> [x]) xs)) : ( insSpaces xss)
insSpaces []       = []

main :: IO ()
main = do
      [f] <- getArgs
      xs <- readFile f
      showSudoku (head (solve (readSudoku xs)))
